from collections.abc import Awaitable
from typing import Any, Literal
from collections.abc import Callable

import numpy as np
import numpy.typing as npt
import zarr.codecs
import zarr.core.config
from zarr.codecs import VLenUTF8Codec
from zarr.core.array import (
    DEFAULT_FILL_VALUE,
    CompressorsLike,
    FiltersLike,
    ShardsLike,
    _parse_keep_array_attr,  # pyright: ignore[reportPrivateUsage]
    init_array,
)
from zarr.core.array_spec import ArrayConfig
from zarr.core.common import JSON, concurrent_map
from zarr.core.dtype import VariableLengthUTF8, ZDType, ZDTypeLike, parse_dtype
from zarr.core.sync import sync
from zarr.errors import ContainsArrayError
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]
from zarr.storage._common import make_store_path  # pyright: ignore[reportUnknownVariableType]

from ezarr.errors import InplaceCastingError


def find_1d(arr: zarr.Array, value: Any) -> int | None:
    assert arr.ndim == 1
    w = np.argwhere(np.array(arr) == value)

    if not w.size:
        return None

    return w[0, 0]


async def _from_array(arr: zarr.Array, dtype: ZDType[np.dtype[np.generic], np.generic]):
    mode: Literal["a"] = "a"
    config = ArrayConfig.from_dict({})
    store_path = await make_store_path(arr.store, path=arr.basename, mode=mode, storage_options=None)

    (
        chunks,
        shards,
        filters,
        compressors,
        serializer,
        fill_value,
        order,
        zarr_format,
        chunk_key_encoding,
        dimension_names,
    ) = _parse_keep_array_attr(
        data=arr,
        chunks="keep",
        shards="keep",
        filters="keep",
        compressors="keep",
        serializer="keep",
        fill_value=DEFAULT_FILL_VALUE,
        order=None,
        zarr_format=None,
        chunk_key_encoding=None,
        dimension_names=None,
    )

    result = await init_array(
        store_path=store_path,
        shape=arr.shape,
        dtype=dtype,
        chunks=chunks,
        shards=shards,
        filters=filters,
        compressors=compressors,
        serializer=serializer,
        fill_value=fill_value,
        order=order,
        zarr_format=zarr_format,
        attributes=None,
        chunk_key_encoding=chunk_key_encoding,
        dimension_names=dimension_names,
        overwrite=True,
        config=config,
    )

    async def _copy_array_region(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
        arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
        await result.setitem(chunk_coords, np.array(arr)[()].astype(dtype.to_native_dtype()))

    # Stream data from the source array to the new array
    await concurrent_map(
        [(region, arr) for region in result._iter_shard_regions()],  # pyright: ignore[reportPrivateUsage]
        _copy_array_region,
        zarr.core.config.config.get("async.concurrency"),
    )

    return result


def cast_array(
    arr: zarr.Array, dtype: ZDTypeLike, casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"]
) -> zarr.Array:
    if arr.dtype == dtype or dtype is None:
        return arr

    zdtype = parse_dtype(dtype, zarr_format=arr.metadata.zarr_format)
    if not np.can_cast(arr.dtype, zdtype.to_native_dtype(), casting):
        raise InplaceCastingError(
            f"Could not cast zarr.Array to type '{zdtype.to_native_dtype()}' according to casting rule '{casting}'"
        )

    return zarr.Array(sync(_from_array(arr, zdtype)))  # pyright: ignore[reportArgumentType]


def write_array(
    arr: zarr.Array | npt.NDArray[np.generic],
    store: StoreLike,
    *,
    name: str,
    dtype: npt.DTypeLike = None,
    casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind",
    chunks: tuple[int, ...] | Literal["auto"] = "auto",
    shards: ShardsLike | None = None,
    filters: FiltersLike = "auto",
    compressors: CompressorsLike = "auto",
    attributes: dict[str, JSON] | None = None,
    overwrite: bool = False,
) -> zarr.Array:
    """
    Write an array to a store while infering the correct dtype.

    Args:
        store: StoreLike
        arr: array to write
        dtype:
        overwrite: overwrite array if it exists (default: False)
    """
    if isinstance(arr, zarr.Array) and arr.store is store and arr.basename == name:
        if not overwrite:
            raise ContainsArrayError(f"An array already exists in store {store} at path '{name}'")

        arr = cast_array(arr, dtype, casting)

        if attributes is not None:
            arr.attrs.put(attributes)

        return arr

    # NOTE: for now, F-style ordering is not handled
    arr = np.ascontiguousarray(arr)

    if dtype is not None:
        arr = arr.astype(dtype, casting=casting)

    if arr.dtype == object:
        if arr.size:
            arr = arr.astype(type(arr.flat[0]))

        if arr.dtype == object:
            raise TypeError(f'Failed interpreting the "Object" data type to a non ambiguous data type for array {arr}')

    if np.issubdtype(arr.dtype, np.str_):
        z_arr = zarr.create_array(
            store,
            name=str(name),
            shape=arr.shape,
            dtype=VariableLengthUTF8(),
            serializer=VLenUTF8Codec(),
            chunks=chunks,
            shards=shards,
            filters=filters,
            compressors=compressors,
            overwrite=overwrite,
            attributes=attributes,
        )
        z_arr[:] = arr
        return z_arr

    return zarr.create_array(
        store,
        name=str(name),
        data=arr,
        chunks=chunks,
        shards=shards,
        filters=filters,
        compressors=compressors,
        overwrite=overwrite,
        attributes=attributes,
    )


async def _concurrent_apply_op(arr: zarr.Array, op: Callable[..., Awaitable[Any]]) -> None:
    await concurrent_map(
        [(region, arr) for region in arr._iter_shard_regions()],  # pyright: ignore[reportPrivateUsage]
        op,
        zarr.core.config.config.get("async.concurrency"),
    )


def apply_op(arr: zarr.Array, op: Callable[..., Awaitable[Any]]) -> None:
    sync(_concurrent_apply_op(arr, op))



import itertools as it
import math
from typing import Any, cast

import numpy as np
import numpy.typing as npt
import zarr

import ezarr


def repr_element(elem: Any, prefix: str = "") -> str:
    if isinstance(elem, np.str_):
        return f"'{elem}'"

    if isinstance(elem, np.generic):
        return str(elem)

    if isinstance(elem, zarr.Group | ezarr.EZDict):
        return "{...}" if len(elem) else "{}"  # pyright: ignore[reportUnknownArgumentType]

    if isinstance(elem, zarr.Array):
        return repr_array(elem, prefix=prefix)

    return repr(elem)


def repr_array(arr: zarr.Array, prefix: str = "") -> str:
    if arr.size == 0:
        return "[]"

    if arr.ndim == 0:
        return repr_element(arr[()])

    data = cast(npt.NDArray[Any], arr.get_orthogonal_selection(tuple(_get_index(shape) for shape in arr.shape)))

    if np.issubdtype(data.dtype, np.integer):
        align = math.ceil(math.log10(data.max()))

    elif np.issubdtype(data.dtype, np.floating):
        data = np.vectorize(np.format_float_positional)(data)
        align = np.vectorize(len)(data).max()

    elif np.issubdtype(data.dtype, np.dtypes.StringDType()):
        align = np.vectorize(len)(data).max()

    else:
        data = data.astype(str)
        align = np.vectorize(len)(data).max()

    return _print3(data, list(arr.shape), align=align, prefix=prefix, ignore_prefix=True)


def _get_index(axis_shape: int) -> list[int]:
    if axis_shape <= 6:
        return list(range(axis_shape))

    return [0, 1, 2, -3, -2, -1]


def _print3(data: np.ndarray, shape: list[int], align: int, *, prefix: str = "", ignore_prefix: bool = False) -> str:
    if data.ndim == 1:
        if shape[0] <= 6:
            values = data
        else:
            values = it.chain(data[:3], ["..."], data[3:])

        return f"{'' if ignore_prefix else prefix}[{' '.join(map(lambda v: f'{v:>{align}}', values))}]"

    if shape[0] <= 6:
        rows = [
            _print3(data[idx], shape[1:], align, prefix=prefix + " ", ignore_prefix=idx == 0) for idx in range(shape[0])
        ]

    else:
        rows = [
            _print3(data[0], shape[1:], align, prefix=prefix + " ", ignore_prefix=True),
            _print3(data[1], shape[1:], align, prefix=prefix + " "),
            _print3(data[2], shape[1:], align, prefix=prefix + " "),
            f"{prefix} ...",
            _print3(data[3], shape[1:], align, prefix=prefix + " "),
            _print3(data[4], shape[1:], align, prefix=prefix + " "),
            _print3(data[5], shape[1:], align, prefix=prefix + " "),
        ]

    sep = "," + "\n" * max(1, len(shape) - 1)
    return f"{'' if ignore_prefix else prefix}[{sep.join(rows)}]"



from collections.abc import Iterable
from types import EllipsisType
from typing import SupportsIndex

from numpy._typing import _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]

type ONE_AXIS_SELECTOR = int | bool | SupportsIndex
type PARTIAL_SELECTOR = None | EllipsisType | slice | range | Iterable[int] | Iterable[bool] | _ArrayLikeInt_co
type SELECTOR = ONE_AXIS_SELECTOR | PARTIAL_SELECTOR



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator, MutableSequence
from pathlib import Path
from typing import Any, Self, cast, overload, override
import warnings

import numpy as np
import numpy.typing as npt
from numpy._typing import _SupportsArray as SupportsArray  # pyright: ignore[reportPrivateUsage]
import zarr
from zarr.core.common import AccessModeLiteral
from zarr.errors import BoundsCheckError, UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

from ezarr._repr import repr_element
from ezarr.names import UNKNOWN, Attribute, EZType
from ezarr.object import EZObject
from ezarr.types import DeferredCreationFunc, ListWrapper, PyObject, PyObjectCodec


def deferred_EZList(values: Collection[Any]) -> DeferredCreationFunc:
    def inner(grp: zarr.Group, name: str, overwrite: bool) -> None:
        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = grp.create_array(
                name,
                shape=(len(values),),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[:] = values  # pyright: ignore[reportArgumentType]

    return inner


class EZList[T](EZObject[list[T]], MutableSequence[T]):
    """
    List-like object built on top of a zarr.Array for storing arbitrary Python objects.

    Args:
        array: a zarr.Array with PyObject() dtype

    Example:
        >>> EZList(zarr.array([1, 2, 3]))
        EZList[1, 2, 3]
    """

    def __init__(self, array: zarr.Array):
        assert array.ndim == 1
        self._array: zarr.Array = array

        self._iter_index: int = 0

    @classmethod
    def from_list(
        cls,
        lst: Collection[T],
        *,
        name: str,
        store: StoreLike | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> Self:
        """
        Create an EZList from a list-like object.

        Args:
            lst: list-like collection of values to store in an EZList.
            name: name for the EZList, to use inside the store.
            store: Store, path to a directory or name of a zip file.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZList will be saved.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> data = [1, "some text", Path("/some/path.txt")]
            >>> ez_list = EZList.from_list(data, name="data")
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt')]

            >>> ez_list = EZList.from_list([1, [2, 3]], name="data")
            >>> ez_list
            EZList[1, array([2, 3])]
        """
        if store is None:
            store = {}

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = zarr.open_group(store, mode=mode, path=path).create_array(
                name,
                shape=(len(lst),),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[:] = np.asarray([ListWrapper(e) if isinstance(e, list | SupportsArray) else e for e in lst])  # pyright: ignore[reportUnknownArgumentType]

        return cls(arr)

    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> npt.NDAarray:
        return self._array[:]

    @override
    def __repr__(self) -> str:
        if len(self) > 6:
            return f"{type(self).__name__}[{repr_element(self[0])}, {repr_element(self[1])}, {repr_element(self[2])}, ..., {repr_element(self[-3])}, {repr_element(self[-2])}, {repr_element(self[-1])}]"

        return f"{type(self).__name__}[{', '.join(repr_element(e) for e in self)}]"

    @override
    def __len__(self) -> int:
        return self._array.shape[0]

    @overload
    def __getitem__(self, item: int) -> T: ...
    @overload
    def __getitem__(self, item: slice[int] | slice[int, int] | slice[int, int, int]) -> Self: ...
    @override
    def __getitem__(self, item: int | slice[int, int, int]) -> T | Self:
        if isinstance(item, slice):
            return type(self)(self._array[item])  # pyright: ignore[reportArgumentType]

        try:
            return cast(T, self._array.get_basic_selection(item)[()])  # pyright: ignore[reportIndexIssue, reportArgumentType, reportCallIssue]

        except BoundsCheckError:
            raise IndexError(
                f"EZList index {item} is out of range [{self._array.shape[0]} ... {self._array.shape[0] - 1}]."
            )

    @overload
    def __setitem__(self, key: int, value: T) -> None: ...
    @overload
    def __setitem__(self, key: slice[int] | slice[int, int] | slice[int, int, int], value: Iterable[T]) -> None: ...
    @override
    def __setitem__(self, key: int | slice[int, int, int], value: T | Iterable[T]) -> None:
        try:
            self._array[key] = np.array(value)

        except BoundsCheckError:
            raise IndexError(
                f"EZList index {key} is out of range [{self._array.shape[0]} ... {self._array.shape[0] - 1}]."
            )

    @override
    def __delitem__(self, key: int | slice[int, int, int]) -> None:
        if isinstance(key, slice):
            raise NotImplementedError

        if key < 0:
            key = len(self) + key

        self._array[key:-1] = self._array[key + 1 :]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.resize(self._array.shape[0] - 1)

    @override
    def __iter__(self) -> Iterator[T]:
        self._iter_index = 0
        return self

    def __next__(self) -> T:
        if self._iter_index >= self._array.shape[0]:
            raise StopIteration

        obj = cast(T, self._array.get_basic_selection(self._iter_index)[()])  # pyright: ignore[reportIndexIssue, reportArgumentType, reportCallIssue]
        self._iter_index += 1

        return obj

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, list | SupportsArray):
            return False

        return np.array_equal(self._array[()], other)  # pyright: ignore[reportUnknownArgumentType]

    @classmethod
    @override
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open this EZList from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZList, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.save("/tmp/list", overwrite=True)
            >>> EZList.open("/tmp/list", name="data")
            EZList[1, 'some text', PosixPath('/some/path.txt')]
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if isinstance(store, str | Path):
            store = Path(store).expanduser()

        arr = zarr.open_array(store, mode=mode, path=path)

        arr_type = cast(str, arr.attrs.get(Attribute.EZType, UNKNOWN))
        if arr_type != EZType.List:
            raise ValueError(f"Cannot read EZList from Array marked with type '{arr_type}', should be 'list'.")

        return cls(arr)

    @override
    def save(
        self,
        store: StoreLike,
        *,
        name: str | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this EZList to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZList, to use inside the store. (default: same as name used during EZLits creation)
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZList will be saved.
            overwrite: overwrite EZList if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.save("/tmp/list", overwrite=True)
            >>> Path("/tmp/list").exists()
            True
        """
        if name is None:
            name = self._array.name

        path = f"{path.rstrip('/')}/{name}" if path else name

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = zarr.create_array(
                store=store,
                name=path,
                shape=(self._array.shape[0],),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[()] = self._array[()]

    @override
    def copy(self) -> list[T]:
        """
        Convert this EZList into a Python list, loading all the data into memory.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.copy()
            [1, 'some text', PosixPath('/some/path.txt')]
        """
        return self._array[:].tolist()  # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue]

    @classmethod
    def defer(cls, values: Collection[Any]) -> DeferredCreationFunc:
        r"""
        Create an EZList from a list-like object, but defer its storage.
        This function returns a function that can be called to finalize the EZList creation and storage,
        with parameters:
            grp: (zarr.Group) a group in which to store the EZList
            name: (str) name for the EZList, to use inside the store.
            overwrite: (bool) overwrite object if a group with name `name` already exists ? (default: False)

        This is usually used as a convenience function to create an EZList in an existing EZDict.

        Args:
            values: list-like collection of values to store in an EZList.

        Example:
            >>> from ezarr import EZDict
            >>> from pathlib import Path

            >>> d = EZDict(zarr.open_group({}))
            >>> d["list"] = EZList.defer([1, 2, 3])
            >>> repr(d)
            'EZDict{\n\tlist: EZList[1, 2, 3]\n}'
        """
        return deferred_EZList(values)

    @override
    def append(self, value: T) -> None:
        """
        Append a value to the end of an EZList.

        Args:
            value: value to be appended.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.append(3.14)
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14]

            >>> ez_list.append([1, 2, 3])
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14, array([1, 2, 3])]
        """
        if isinstance(value, list | SupportsArray):
            value = ListWrapper(value)  # pyright: ignore[reportAssignmentType, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.append(np.asarray([value]))

    @override
    def extend(self, values: Iterable[T]) -> None:
        """
        Append multiple values to the end of an EZList.

        Args:
            values: values to be appended.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.extend([3.14, 2.71, [-1, -2]])
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14, 2.71, array([-1, -2])]
        """
        values_np = np.asarray([ListWrapper(v) if isinstance(v, list | SupportsArray) else v for v in values])  # pyright: ignore[, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.append(values_np)

    @override
    def insert(self, index: int, value: T) -> None:
        """
        Insert a value in an EZList before `index`.

        Args:
            value: value to be inserted.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.insert(1, [1, 2, 3])
            >>> ez_list
            EZList[1, array([1, 2, 3]), 'some text', PosixPath('/some/path.txt')]
        """
        if index < 0:
            index = len(self) + index

        if isinstance(value, list | SupportsArray):
            value = ListWrapper(value)  # pyright: ignore[reportAssignmentType, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.resize(self._array.shape[0] + 1)
        self._array[index + 1 :] = self._array[index:-1]

        self._array[index] = value  # pyright: ignore[reportArgumentType]



import numpy as np
import pandas as pd

from ezarr.dataframe import EZDataFrame


def test_create_empty() -> None:
    ezdf = EZDataFrame()
    assert (
        repr(ezdf)
        == """\
Empty EZDataFrame
Columns: []
Index: []
[RAM]"""
    )


def test_can_create_from_pandas_DataFrame(ezdf: EZDataFrame) -> None:
    assert isinstance(ezdf, EZDataFrame)


def test_has_correct_columns(ezdf: EZDataFrame) -> None:
    assert np.array_equal(ezdf.columns, ["col_int", "col_str", "col_float"])


def test_has_correct_repr(ezdf: EZDataFrame) -> None:
    assert (
        repr(ezdf)
        == """\
   col_int col_str  col_float
0        1       a        1.5
1        2      bc        2.5
2        3     def        3.5
[Local]
[3 rows x 3 columns]"""
    )


def test_convert_to_pandas(ezdf: EZDataFrame) -> None:
    df = ezdf.copy()

    assert isinstance(df, pd.DataFrame)
    assert np.array_equal(df.index, [0, 1, 2])
    assert np.array_equal(df.columns, ["col_int", "col_str", "col_float"])
    assert np.array_equal(df.col_int, [1, 2, 3])
    assert np.array_equal(df.col_str, ["a", "bc", "def"])


def test_can_get_column_from_getattr(ezdf: EZDataFrame) -> None:
    assert np.array_equal(ezdf.col_str, ["a", "bc", "def"])


def test_can_set_existing_column(ezdf: EZDataFrame) -> None:
    ezdf["col_int"] = -1
    assert np.array_equal(ezdf["col_int"], [-1, -1, -1])


def test_can_set_new_column_str(ezdf: EZDataFrame) -> None:
    ezdf["new"] = "new!"
    assert np.array_equal(ezdf["new"], ["new!", "new!", "new!"])


def test_can_set_new_column_int(ezdf: EZDataFrame) -> None:
    ezdf["new"] = [-1, -2, -3]
    assert np.array_equal(ezdf["new"], [-1, -2, -3])


def test_can_set_column_no_index() -> None:
    ezdf = EZDataFrame(pd.DataFrame(columns=["value"], dtype=int))
    ezdf["value"] = [1, 2, 3]

    assert np.array_equal(ezdf["value"], [1, 2, 3])


def test_set_value_sets_in_h5_file(ezdf: EZDataFrame) -> None:
    ezdf.loc[1, "col_str"] = "test"

    assert np.array_equal(ezdf._data_store["arrays"]["col_str"], ["a", "test", "def"])



import warnings
from typing import Any, Literal, SupportsIndex, cast

import numpy as np
import numpy.typing as npt
import zarr
from numpy._typing import ArrayLike, _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]
from zarr.errors import UnstableSpecificationWarning

from ezarr.common import apply_op, cast_array
from ezarr.view import ArrayView


def _check_axis(arr: zarr.Array, axis: SupportsIndex | None) -> int:
    if arr.ndim > 1 and axis is None:
        raise NotImplementedError("Cannot flatten zarr.Array while working inplace")

    if axis is None:
        return 0

    return int(axis)


def _clamp_index(arr: zarr.Array, axis: SupportsIndex, obj: _ArrayLikeInt_co | slice) -> np.integer:
    if not isinstance(obj, int | np.integer):
        raise NotImplementedError("Advanced indexing is not supported yet")

    obj = cast(np.integer, obj)

    if obj > arr.shape[axis]:
        raise IndexError(f"Index {obj} is out of bounds for axis {axis} with size {arr.shape[axis]}.")

    elif obj < 0:
        obj = cast(np.integer, obj + arr.shape[axis])

    return obj


def delete[A: zarr.Array | npt.NDArray[Any]](
    arr: A, obj: _ArrayLikeInt_co | slice, axis: SupportsIndex | None = None
) -> A:
    if not isinstance(arr, zarr.Array):
        return np.delete(arr, obj, axis)

    axis = _check_axis(arr, axis)
    obj = _clamp_index(arr, axis, obj)

    prefix = (slice(None),) * axis
    if obj < (arr.shape[axis] - 1):
        # transfer data one row to the left, starting from the column after the one to delete
        # matrix | 0 1 2 3 4 | with index of the column to delete = 2
        #   ==>  | 0 1 3 4 . |
        index_dest = prefix + (slice(obj, -1),)
        index_source = prefix + (slice(obj + 1, None),)
        arr[index_dest] = arr[index_source]

    # resize the arrays to drop the extra column at the end
    # matrix | 0 1 3 4 . |
    #   ==>  | 0 1 3 4 |
    new_shape = arr.shape[:axis] + (arr.shape[axis] - 1,) + arr.shape[axis + 1 :]

    with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
        arr.resize(new_shape)

    return arr


def insert[A: zarr.Array | npt.NDArray[Any]](
    arr: A, obj: _ArrayLikeInt_co | slice, values: ArrayLike, axis: SupportsIndex | None = None
) -> A:
    """
    Insert values along the given axis before the given indices.

    Args:
        arr: input zarr.Array.
        obj: object that defines the index or indices before which `values` is inserted.
        values: values to inset into `arr`. If the type of `values` is different from that of `arr`, `values` is
            converted to the type of `arr`. `values` should be shaped so that `arr[...,obj,...] = values` is legal.
        axis: (optional) axis along which to insert `values`. If axis is None then arr is flattened first.
    """
    if not isinstance(arr, zarr.Array):
        return np.insert(arr, obj, values, axis)

    axis = _check_axis(arr, axis)
    obj = _clamp_index(arr, axis, obj)

    # resize the arrays to make room for the extra values
    new_shape = arr.shape[:axis] + (arr.shape[axis] + 1,) + arr.shape[axis + 1 :]

    with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
        arr.resize(new_shape)

    prefix = (slice(None),) * axis
    if obj < (arr.shape[axis] - 1):
        # transfer data 1 row to the right, starting from the column after the insertion index
        # matrix | 0 1 2 3 4 | with index of the insertion column = 2
        #   ==>  | 0 1 . 2 3 4 |
        index_source = prefix + (slice(obj, -1),)
        index_dest = prefix + (slice(obj + 1, None),)
        arr[index_dest] = arr[index_source]

    arr.set_orthogonal_selection(prefix + (obj,), values)  # pyright: ignore[reportArgumentType]

    return arr


def add[A: zarr.Array | npt.NDArray[Any] | ArrayView[Any]](
    x1: A, x2: ArrayLike, /, casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind"
) -> A:
    if not isinstance(x1, zarr.Array):
        if isinstance(x1, ArrayView):
            x1 += x2
            return x1

        return np.add(x1, x2)

    is_scalar = np.isscalar(x2)

    x2 = np.asarray(x2)
    x1 = cast_array(x1, np.result_type(x1.dtype, x2.dtype), casting)

    if is_scalar:

        async def _add_scalar(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr + x2)  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_scalar)

    else:

        async def _add_array(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr + x2[chunk_coords])  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_array)

    return x1


def sub[A: zarr.Array | npt.NDArray[Any] | ArrayView[Any]](
    x1: A, x2: ArrayLike, /, casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind"
) -> A:
    if not isinstance(x1, zarr.Array):
        if isinstance(x1, ArrayView):
            x1 -= x2
            return x1

        return np.subtract(x1, x2)

    is_scalar = np.isscalar(x2)

    x2 = np.asarray(x2)
    x1 = cast_array(x1, np.result_type(x1.dtype, x2.dtype), casting)

    if is_scalar:

        async def _add_scalar(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr - x2)  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_scalar)

    else:

        async def _add_array(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr - x2[chunk_coords])  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_array)

    return x1


def multiply[A: zarr.Array | npt.NDArray[Any] | ArrayView[Any]](
    x1: A, x2: ArrayLike, /, casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind"
) -> A:
    if not isinstance(x1, zarr.Array):
        if isinstance(x1, ArrayView):
            x1 *= x2
            return x1

        return np.multiply(x1, x2)

    is_scalar = np.isscalar(x2)

    x2 = np.asarray(x2)
    x1 = cast_array(x1, np.result_type(x1.dtype, x2.dtype), casting)

    if is_scalar:

        async def _add_scalar(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr * x2)  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_scalar)

    else:

        async def _add_array(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr * x2[chunk_coords])  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_array)

    return x1


def true_divide[A: zarr.Array | npt.NDArray[Any] | ArrayView[Any]](
    x1: A, x2: ArrayLike, /, casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] = "same_kind"
) -> A:
    if not isinstance(x1, zarr.Array):
        if isinstance(x1, ArrayView):
            x1 /= x2
            return x1

        return np.divide(x1, x2)

    is_scalar = np.isscalar(x2)

    x2 = np.asarray(x2)
    x1 = cast_array(x1, np.result_type(x1.dtype, x2.dtype), casting)

    if is_scalar:

        async def _add_scalar(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr / x2)  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_scalar)

    else:

        async def _add_array(chunk_coords: tuple[int, ...] | slice, _data: zarr.Array) -> None:
            arr = await _data._async_array.getitem(chunk_coords)  # pyright: ignore[reportPrivateUsage]
            await _data._async_array.setitem(chunk_coords, arr / x2[chunk_coords])  # pyright: ignore[reportPrivateUsage]

        apply_op(x1, _add_array)

    return x1



from typing import Any, Callable, Hashable, Literal, Protocol, Self, override

import numpy as np
import numpy.typing as npt
from pandas._typing import ArrayLike, AxisInt, FloatFormatType
from pandas.core.arrays import ExtensionArray
from pandas.core.indexes.api import Index
from pandas.core.internals.base import DataManager, SingleDataManager

class ArrayApplyFunc[R](Protocol):
    def __call__(self, arr: npt.NDArray[Any], **kwds: Any) -> R: ...

class BaseArrayManager(DataManager):
    """
    Core internal data structure to implement DataFrame and Series.

    Alternative to the BlockManager, storing a list of 1D arrays instead of
    Blocks.

    This is *not* a public API class

    Parameters
    ----------
    arrays : Sequence of arrays
    axes : Sequence of Index
    verify_integrity : bool, default True

    """

    __slots__: list[str] = ["_axes", "arrays"]

    arrays: list[np.ndarray | ExtensionArray]
    _axes: list[Index]

    def __init__(
        self, arrays: list[np.ndarray | ExtensionArray], axes: list[Index], verify_integrity: bool = True
    ) -> None: ...
    def make_empty(self, axes: list[Index] | None = None) -> Self: ...
    @property
    def items(self) -> Index: ...
    @property
    def axes(self) -> list[Index]: ...
    @property
    def shape_proper(self) -> tuple[int, ...]: ...
    @staticmethod
    def _normalize_axis(axis: AxisInt) -> int: ...
    def set_axis(self, axis: AxisInt, new_labels: Index) -> None: ...
    def get_dtypes(self) -> npt.NDArray[np.object_]: ...
    def add_references(self, mgr: BaseArrayManager) -> None: ...
    @override
    def __getstate__(self) -> tuple[list[npt.NDArray[Any] | ExtensionArray], list[Index]]: ...
    def __setstate__(self, state: tuple[list[npt.NDArray[Any] | ExtensionArray], list[Index]]) -> None: ...
    @override
    def __repr__(self) -> str: ...
    def apply(self, f: str | ArrayApplyFunc[ArrayLike], align_keys: list[str] | None = None, **kwargs: Any) -> Self: ...
    def apply_with_block(self, f: str, align_keys: list[str] | None = None, **kwargs: Any) -> Self: ...
    def setitem(self, indexer: Any, value: Any, warn: bool = True) -> Self: ...
    def diff(self, n: int) -> Self: ...
    def astype(self, dtype: npt.DTypeLike, copy: bool | None = False, errors: str = "raise") -> Self: ...
    def convert(self, copy: bool | None) -> Self: ...
    def get_values_for_csv(
        self,
        *,
        float_format: FloatFormatType,
        date_format: Any,
        decimal: str,
        na_rep: str = "nan",
        quoting: int | None = None,
    ) -> Self: ...
    @property
    def any_extension_types(self) -> bool: ...
    @property
    def is_view(self) -> bool: ...
    @property
    def is_single_block(self) -> bool: ...
    def _get_data_subset(self, predicate: Callable[[npt.NDArray[Any]], bool]) -> Self: ...
    def get_bool_data(self, copy: bool = False) -> Self: ...
    def get_numeric_data(self, copy: bool = False) -> Self: ...
    def copy(self, deep: bool | Literal["all"] | None = True) -> Self: ...
    def reindex_indexer(
        self,
        new_axis,
        indexer,
        axis: AxisInt,
        fill_value=None,
        allow_dups: bool = False,
        copy: bool | None = True,
        only_slice: bool = False,
        use_na_proxy: bool = False,
    ) -> Self: ...
    def _reindex_indexer(
        self,
        new_axis,
        indexer: npt.NDArray[np.intp] | None,
        axis: AxisInt,
        fill_value=None,
        allow_dups: bool = False,
        copy: bool | None = True,
        use_na_proxy: bool = False,
    ) -> Self: ...
    def take(self, indexer: npt.NDArray[np.intp], axis: AxisInt = 1, verify: bool = True) -> Self: ...
    def _make_na_array(self, fill_value=None, use_na_proxy: bool = False): ...
    def _equal_values(self, other) -> bool: ...

class ArrayManager(BaseArrayManager):
    @property
    def ndim(self) -> Literal[2]: ...
    def __init__(
        self, arrays: list[np.ndarray | ExtensionArray], axes: list[Index], verify_integrity: bool = True
    ) -> None: ...
    def _verify_integrity(self) -> None: ...
    def fast_xs(self, loc: int) -> SingleArrayManager: ...
    def get_slice(self, slobj: slice, axis: AxisInt = 0) -> ArrayManager: ...
    def iget(self, i: int) -> SingleArrayManager: ...
    def iget_values(self, i: int) -> ArrayLike: ...
    @property
    def column_arrays(self) -> list[ArrayLike]: ...
    def iset(self, loc: int | slice | np.ndarray, value: ArrayLike, inplace: bool = False, refs=None) -> None: ...
    def column_setitem(self, loc: int, idx: int | slice | np.ndarray, value, inplace_only: bool = False) -> None: ...
    def insert(self, loc: int, item: Hashable, value: ArrayLike, refs=None) -> None: ...
    def idelete(self, indexer) -> ArrayManager: ...
    def grouped_reduce(self, func: Callable) -> Self: ...
    def reduce(self, func: Callable) -> Self: ...
    def operate_blockwise(self, other: ArrayManager, array_op) -> ArrayManager: ...
    def quantile(
        self, *, qs: Index, transposed: bool = False, interpolation: QuantileInterpolation = "linear"
    ) -> ArrayManager: ...
    def unstack(self, unstacker, fill_value) -> ArrayManager: ...
    def as_array(self, dtype=None, copy: bool = False, na_value: object = lib.no_default) -> np.ndarray: ...
    @classmethod
    def concat_horizontal(cls, mgrs: list[Self], axes: list[Index]) -> Self: ...
    @classmethod
    def concat_vertical(cls, mgrs: list[Self], axes: list[Index]) -> Self: ...

class SingleArrayManager(BaseArrayManager, SingleDataManager):
    __slots__: list[str] = ["_axes", "arrays"]

    arrays: list[np.ndarray | ExtensionArray]
    _axes: list[Index]

    @property
    def ndim(self) -> Literal[1]: ...
    def __init__(
        self, arrays: list[np.ndarray | ExtensionArray], axes: list[Index], verify_integrity: bool = True
    ) -> None: ...
    def _verify_integrity(self) -> None: ...
    @staticmethod
    @override
    def _normalize_axis(axis: AxisInt) -> AxisInt: ...
    @override
    def make_empty(self, axes: list[Index] | None = None) -> Self: ...
    @classmethod
    def from_array(cls, array: npt.NDArray[Any] | ExtensionArray, index: Index) -> SingleArrayManager: ...
    @property
    @override
    def axes(self) -> list[Index]: ...
    @property
    def index(self) -> Index: ...
    @property
    def dtype(self): ...
    def external_values(self): ...
    def internal_values(self): ...
    def array_values(self): ...
    @property
    def _can_hold_na(self) -> bool: ...
    @property
    @override
    def is_single_block(self) -> bool: ...
    def fast_xs(self, loc: int) -> SingleArrayManager: ...
    def get_slice(self, slobj: slice, axis: AxisInt = 0) -> SingleArrayManager: ...
    def get_rows_with_mask(self, indexer: npt.NDArray[np.bool_]) -> SingleArrayManager: ...
    @override
    def apply(self, func: str | Callable[[Any], npt.NDArray[Any] | ExtensionArray], **kwargs: Any) -> Self: ...  # pyright: ignore[reportIncompatibleMethodOverride]
    @override
    def setitem(self, indexer: npt.NDArray[np.integer], value: Any, warn: bool = True) -> SingleArrayManager: ...
    def idelete(self, indexer: npt.NDArray[np.integer]) -> SingleArrayManager: ...
    @override
    def _get_data_subset(self, predicate: Callable[[npt.NDArray[Any]], bool]) -> SingleArrayManager: ...
    def set_values(self, values: ArrayLike) -> None: ...
    def to_2d_mgr(self, columns: Index) -> ArrayManager: ...



from typing import Literal
import numpy as np
import pandas as pd
import pytest

from ezarr.dataframe import EZDataFrame


def test_can_map_function(ezdf: EZDataFrame) -> None:
    res = ezdf.col_int.map(lambda x: x**2)

    assert np.array_equal(res, [1, 4, 9])


@pytest.mark.parametrize("klass", ["pandas", "ezarr"])
def test_can_test_equality(ezdf: EZDataFrame, klass: Literal["pandas", "ezarr"]) -> None:
    other = pd.DataFrame({"col_int": [1, 2, 3], "col_str": ["a", "bc", "def"], "col_float": [1.5, 2.5, 3.5]})

    if klass == "ezarr":
        other = EZDataFrame(other)

    assert np.array_equal(ezdf, other)



from __future__ import annotations

from typing import Any, Self, override

import numpy as np
import pytest
import zarr

import ezarr
from ezarr import io
from ezarr.types import SupportsEZReadWrite


class O:  # noqa: E742
    """A Python object"""

    def __init__(self, v: float):
        self._v: float = v

    @override
    def __repr__(self) -> str:
        return f"O({self._v})"

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, O):
            return False

        return self._v == other._v


class P(SupportsEZReadWrite):
    """A Python object that is a SupportsEZReadWrite"""

    def __init__(self, l_: list[int]):
        self.l: list[int] = l_

    @override
    def __repr__(self) -> str:
        return f"P([{', '.join(map(str, self.l))}])"

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, P):
            return False

        return self.l == other.l

    @override
    @classmethod
    def __ez_read__(cls, values: ezarr.EZDict[Any]) -> Self:
        arr = values["l"]
        assert isinstance(arr, zarr.Array)

        return cls(list(arr[:].tolist()))  # pyright: ignore[reportUnknownArgumentType, reportAttributeAccessIssue]

    @override
    def __ez_write__(self, values: ezarr.EZDict[Any]) -> None:
        io.write_object(values, obj=self.l, name="l")


@pytest.fixture
def ez_list() -> ezarr.EZList[Any]:
    data = [1.0, 2, P([1, 2, 3]), "4.", O(5.0)]

    return ezarr.EZList[Any].from_list(data, store={}, name="data")


def test_list_repr(ez_list: ezarr.EZList[Any]):
    assert repr(ez_list) == "EZList[1.0, 2, P([1, 2, 3]), '4.', O(5.0)]"


def test_list_should_read_custom_object(ez_list: ezarr.EZList[Any]):
    assert isinstance(ez_list[4], O)


def test_list_should_read_custom_object_with_method(ez_list: ezarr.EZList[Any]):
    assert isinstance(ez_list[2], P)


def test_list_should_convert_to_regular_list(ez_list: ezarr.EZList[Any]):
    lst = ez_list.copy()
    assert lst == [1.0, 2, P([1, 2, 3]), "4.", O(5.0)]


def test_list_get_negative_index(ez_list: ezarr.EZList[Any]):
    assert ez_list[-2] == "4."


def test_list_can_append_value_to_list(ez_list: ezarr.EZList[Any]):
    ez_list.append(-1)
    assert ez_list[5] == -1


def test_list_can_append_array_to_list(ez_list: ezarr.EZList[Any]):
    ez_list.append([7, 8, 9])
    assert np.array_equal(ez_list[5], [7, 8, 9])


def test_list_can_create_deferred_list(ez_dict: ezarr.EZDict[Any]):
    ez_dict["z"] = ezarr.EZList.defer([])

    assert isinstance(ez_dict["z"], ezarr.EZList)


def test_list_can_create_deferred_list_with_values(ez_dict: ezarr.EZDict[Any]):
    ez_dict["zz"] = ezarr.EZList.defer([1, 2, 3])

    assert isinstance(ez_dict["zz"], ezarr.EZList)
    assert ez_dict["zz"] == [1, 2, 3]


def test_list_can_delete_last_item(ez_list: ezarr.EZList[Any]):
    del ez_list[-1]
    assert ez_list == [1.0, 2, P([1, 2, 3]), "4."]


def test_list_can_delete_from_middle(ez_list: ezarr.EZList[Any]):
    del ez_list[2]
    assert ez_list == [1.0, 2, "4.", O(5.0)]



import numpy as np
import pandas as pd
import pytest
import zarr

from ezarr.dataframe import EZDataFrame


def test_can_write() -> None:
    ezdf = EZDataFrame(
        pd.DataFrame(
            {"col_int": [1, 2, 3], "col_str": ["a", "bc", "def"], "col_float": [1.5, 2.5, 3.5]}, index=["a", "b", "c"]
        )
    )

    store = zarr.create_group({})
    ezdf.save(store.store, name="data")

    assert set(store["data"].keys()) == {"index", "arrays"}  # pyright: ignore[reportUnknownArgumentType, reportAttributeAccessIssue]
    assert np.array_equal(store["data/index"], ["a", "b", "c"])  # pyright: ignore[reportArgumentType]
    assert set(store["data/arrays"].keys()) == {"col_float", "col_int", "col_str"}  # pyright: ignore[reportUnknownArgumentType, reportAttributeAccessIssue]
    assert np.array_equal(store["data/arrays/col_int"], [1, 2, 3])  # pyright: ignore[reportArgumentType]


@pytest.mark.filterwarnings("ignore")
def test_can_read() -> None:
    store = zarr.create_group({})

    store["data/index"] = np.array(["a", "b", "c"])
    store["data/arrays/col_int"] = np.array([1, 2, 3])
    store["data/arrays/col_str"] = np.array(["a", "bc", "def"])
    store["data/arrays"].attrs["columns_order"] = ["col_int", "col_str"]

    ezdf = EZDataFrame.open(store.store, name="data")
    assert np.array_equal(ezdf.index, ["a", "b", "c"])
    assert np.array_equal(ezdf.columns, ["col_int", "col_str"])



from __future__ import annotations

import itertools
import warnings
from collections.abc import Callable, Hashable
from typing import Any, Literal, cast, override

import numpy as np
import numpy.typing as npt
import pandas as pd
import zarr
from pandas._libs import lib
from pandas._typing import QuantileInterpolation
from pandas.core.array_algos.quantile import quantile_compat  # pyright: ignore[reportUnknownVariableType, reportMissingImports]
from pandas.core.array_algos.take import take_1d  # pyright: ignore[reportUnknownVariableType, reportMissingImports]
from pandas.core.arrays import ExtensionArray
from pandas.core.construction import extract_array, sanitize_array  # pyright: ignore[reportAttributeAccessIssue, reportUnknownVariableType]
from pandas.core.dtypes.cast import np_can_hold_element  # pyright: ignore[reportAttributeAccessIssue, reportUnknownVariableType]
from pandas.core.dtypes.common import ensure_platform_int  # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue]
from pandas.core.dtypes.missing import isna
from pandas.core.indexers.utils import validate_indices  # pyright: ignore[reportUnknownVariableType, reportMissingImports]
from pandas.core.indexes.base import ensure_index  # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue]
from pandas.core.internals.array_manager import (  # pyright: ignore[reportMissingImports]
    ArrayManager,  # pyright: ignore[reportUnknownVariableType]
    DataManager,  # pyright: ignore[reportUnknownVariableType]
    SingleArrayManager,  # pyright: ignore[reportUnknownVariableType]
    concat_arrays,  # pyright: ignore[reportUnknownVariableType]
)
from pandas.core.internals.base import ensure_np_dtype, interleaved_dtype  # pyright: ignore[reportMissingImports, reportUnknownVariableType]
from pandas.core.internals.blocks import ensure_block_shape, external_values  # pyright: ignore[reportMissingImports, reportUnknownVariableType]
from pandas.core.internals.managers import BlockManager  # pyright: ignore[reportMissingImports, reportUnknownVariableType]
from zarr.errors import UnstableSpecificationWarning  # type: ignore[import-untyped]


def extract_pandas_array(array: Any) -> zarr.Array:
    if not isinstance(array, zarr.Array):
        raise TypeError(f"Only zarr.Arrays are allowed in EZDataFrames, got '{type(array)}'")

    return array


def array_equals(left: npt.NDArray[Any], right: npt.NDArray[Any]) -> bool:
    """
    ExtensionArray-compatible implementation of array_equivalent.
    """
    if left.dtype != right.dtype and (left.dtype.kind not in "OSUT" or right.dtype.kind not in "OSUT"):
        return False

    return np.array_equal(left, right)


class EZArrayManager(ArrayManager):  # pyright: ignore[reportUntypedBaseClass]
    @override
    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self,
        arrays: list[zarr.Array],
        axes: list[pd.Index],
        group: zarr.Group | None = None,
        verify_integrity: bool = True,
    ) -> None:
        if verify_integrity:
            self._axes: list[pd.Index] = [ensure_index(ax) for ax in axes]
            self.arrays: list[zarr.Array] = [extract_pandas_array(x) for x in arrays]
            self._verify_integrity()

        else:
            # Note: we are storing the axes in "_axes" in the (row, columns) order
            # which contrasts the order how it is stored in BlockManager
            self._axes = axes
            self.arrays = [extract_pandas_array(x) for x in arrays]

        if group is None:
            if not len(self.arrays):
                raise ValueError("Cannot create EZArrayManager with no group and no arrays")

            arr = self.arrays[0]
            group = zarr.open_group(arr.store, path=arr.name.rstrip(arr.basename))

        self._group: zarr.Group = group

    def _verify_integrity(self) -> None:
        n_columns, n_rows = cast(tuple[int, int], self.shape)

        if not len(self.arrays) == n_columns:
            raise ValueError(
                f"Number of passed arrays must equal the size of the column Index: {len(self.arrays)} arrays vs {n_columns} columns."
            )

        for arr in self.arrays:
            if not arr.shape[0] == n_rows:
                raise ValueError(
                    f"Passed arrays should have the same length as the rows Index: {arr.shape[0]} vs {n_rows} rows"
                )

            if not arr.ndim == 1:
                raise ValueError(
                    f"Passed arrays should be 1-dimensional, got array with {arr.ndim} dimensions instead."
                )

    @property
    def ndim(self) -> Literal[2]:
        return 2

    # --------------------------------------------------------------------
    # Indexing
    def fast_xs(self, loc: int) -> SingleArrayManager:
        """
        Return the array corresponding to `frame.iloc[loc]`.

        Parameters
        ----------
        loc : int

        Returns
        -------
        np.ndarray or ExtensionArray
        """
        dtype = interleaved_dtype([arr.dtype for arr in self.arrays])

        values = [arr[loc] for arr in self.arrays]
        result = np.hstack(values, dtype=dtype)
        return SingleArrayManager([result], [self._axes[1]])

    def get_slice(self, slobj: slice, axis: int = 0) -> ArrayManager:
        axis = self._normalize_axis(axis)

        if axis == 0:
            arrays = [arr[slobj] for arr in self.arrays]

        elif axis == 1:
            arrays = self.arrays[slobj]

        else:
            raise ValueError(f"Invalid axis: {axis}")

        new_axes = list(self._axes)
        new_axes[axis] = new_axes[axis]._getitem_slice(slobj)  # type: ignore[attr-defined]

        return ArrayManager(arrays, new_axes, verify_integrity=False)

    def iget(self, i: int) -> EZSingleArrayManager:
        """
        Return the data as a H5SingleArrayManager.
        """
        return EZSingleArrayManager([self.arrays[i]], [self._axes[0]])

    def iget_values(self, i: int) -> zarr.Array:
        """
        Return the data for column i as the values (ndarray or ExtensionArray).
        """
        return self.arrays[i]

    @property
    def column_arrays(self) -> list[npt.NDArray[Any]]:
        """
        Used in the JSON C code to access column arrays.
        """
        return [np.asarray(arr) for arr in self.arrays]

    def iset(
        self,
        loc: int | slice | npt.NDArray[np.bool_],
        value: npt.NDArray[Any] | ExtensionArray,
        inplace: bool = False,
        refs: None = None,
    ) -> None:
        """
        Set new column(s).

        This changes the ArrayManager in-place, but replaces (an) existing
        column(s), not changing column values in-place).

        Parameters
        ----------
        loc : integer, slice or boolean mask
            Positional location (already bounds checked)
        value : np.ndarray or ExtensionArray
        inplace : bool, default False
            Whether overwrite existing array as opposed to replacing it.
        """
        # single column -> single integer index
        if lib.is_integer(loc):
            # TODO can we avoid needing to unpack this here? That means converting
            # DataFrame into 1D array when loc is an integer
            if isinstance(value, np.ndarray) and value.ndim == 2:
                assert value.shape[1] == 1
                value = value[:, 0]

            assert isinstance(value, np.ndarray)
            assert value.ndim == 1
            assert len(value) == len(self._axes[0])

            self.arrays[loc][:] = value
            return

        # multiple columns -> convert slice or array to integer indices
        elif isinstance(loc, slice):
            indices: range | npt.NDArray[Any] = range(
                loc.start if loc.start is not None else 0,
                loc.stop if loc.stop is not None else self.shape_proper[1],
                loc.step if loc.step is not None else 1,
            )

        else:
            assert isinstance(loc, np.ndarray)
            assert loc.dtype == "bool"
            indices = np.nonzero(loc)[0]

        assert value.ndim == 2
        assert value.shape[0] == len(self._axes[0])

        for value_idx, mgr_idx in enumerate(indices):
            assert isinstance(mgr_idx, int | np.integer)

            value_arr = value[:, value_idx]  # pyright: ignore[reportCallIssue, reportArgumentType]
            self.arrays[mgr_idx][:] = value_arr

    def column_setitem(
        self, loc: int, idx: int | slice | npt.NDArray[Any], value: Any, inplace_only: bool = False
    ) -> None:
        """
        Set values ("setitem") into a single column (not setting the full column).

        This is a method on the ArrayManager level, to avoid creating an
        intermediate Series at the DataFrame level (`s = df[loc]; s[idx] = value`)
        """
        if not lib.is_integer(loc):
            raise TypeError("The column index should be an integer")

        arr = self.arrays[loc]
        mgr = EZSingleArrayManager([arr], [self._axes[0]])

        mgr.setitem_inplace(idx, value)

    def insert(self, loc: int, item: Hashable, value: npt.NDArray[Any], refs: None = None) -> None:
        """
        Insert item at selected position.

        Parameters
        ----------
        loc : int
        item : hashable
        value : np.ndarray or ExtensionArray
        """
        value = extract_array(value, extract_numpy=True)
        if value.ndim == 2:
            if value.shape[0] == 1:
                value = value[0, :]

            else:
                raise ValueError(f"Expected a 1D array, got an array with shape {value.shape}")

        if value.dtype == object:
            value = value.astype(str)

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            z_value = self._group.require_group("arrays").create_array(str(item), data=value)
            self._group["arrays"].attrs["columns_order"] = self._group["arrays"].attrs.setdefault(
                "columns_order", []
            ) + [str(item)]  # pyright: ignore[reportOperatorIssue]

        self.arrays.insert(loc, z_value)
        self._axes[1] = self.items.insert(loc, item)

    def idelete(self, indexer: npt.NDArray[np.intp] | None) -> EZArrayManager:
        """
        Delete selected locations in-place (new block and array, same BlockManager)
        """
        to_keep = np.ones(self.shape[0], dtype=np.bool_)
        to_keep[indexer] = False

        self.arrays = [self.arrays[i] for i in np.nonzero(to_keep)[0]]
        self._axes = [self._axes[0], self._axes[1][to_keep]]
        return self

    def _reindex_indexer(
        self,
        new_axis: pd.Index,
        indexer: npt.NDArray[np.intp] | None,
        axis: int,
        fill_value: None = None,
        allow_dups: bool = False,
        copy: bool | None = True,
        use_na_proxy: bool = False,
    ) -> ArrayManager:
        """
        Parameters
        ----------
        new_axis : Index
        indexer : ndarray[intp] or None
        axis : int
        fill_value : object, default None
        allow_dups : bool, default False
        copy : bool, default True


        pandas-indexer with -1's only.
        """
        if copy is None:
            # ArrayManager does not yet support CoW, so deep=None always means
            # deep=True for now
            copy = True

        if indexer is None:
            if new_axis is self._axes[axis] and not copy:
                return self

            new_axes = list(self._axes)
            new_axes[axis] = new_axis
            return type(self)(self.arrays, new_axes, group=self._group, verify_integrity=False)

        # some axes don't allow reindexing with dups
        if not allow_dups:
            self._axes[axis]._validate_can_reindex(indexer)  # type: ignore[attr-defined]

        if axis >= self.ndim:
            raise IndexError("Requested axis not found in manager")

        new_axes = list(self._axes)
        new_axes[axis] = new_axis

        if axis == 1:
            new_arrays = []
            for i in indexer:
                if i == -1:
                    raise NotImplementedError
                    arr = self._make_na_array(fill_value=fill_value, use_na_proxy=use_na_proxy)

                else:
                    assert copy

                    arr = self.arrays[i]
                    if copy:
                        arr = arr.copy()

                new_arrays.append(arr)

            if copy:
                return ArrayManager(new_arrays, new_axes, verify_integrity=False)

        else:
            validate_indices(indexer, len(self._axes[0]))
            indexer = ensure_platform_int(indexer)
            mask = cast(npt.NDArray[np.bool_], indexer == -1)
            needs_masking = mask.any()

            if copy:
                new_arrays = [
                    take_1d(
                        arr[:],
                        indexer,
                        allow_fill=needs_masking,
                        fill_value="" if isinstance(arr.dtype, np.dtypes.StringDType) else fill_value,
                        mask=mask,
                    )
                    for arr in self.arrays
                ]

                return ArrayManager(new_arrays, new_axes, verify_integrity=False)

            for idx, arr in enumerate(self.arrays):
                new_arr = take_1d(
                    arr[:],
                    indexer,
                    allow_fill=needs_masking,
                    fill_value="" if isinstance(arr.dtype, np.dtypes.StringDType) else fill_value,
                    mask=mask,
                )
                self.arrays[idx] = self._group.create_array(
                    arr.basename, data=new_arr.astype(arr.dtype), overwrite=True
                )

        return type(self)(self.arrays, new_axes, group=self._group, verify_integrity=False)

    # --------------------------------------------------------------------
    # Array-wise Operation
    def grouped_reduce(self, func: Callable[..., Any]) -> EZArrayManager:
        """
        Apply grouped reduction function columnwise, returning a new ArrayManager.

        Parameters
        ----------
        func : grouped reduction function

        Returns
        -------
        ArrayManager
        """
        result_arrays: list[npt.NDArray[Any]] = []
        result_indices: list[int] = []

        for i, arr in enumerate(self.arrays):
            # grouped_reduce functions all expect 2D arrays
            arr = ensure_block_shape(arr, ndim=2)
            res = func(arr)
            if res.ndim == 2:
                # reverse of ensure_block_shape
                assert res.shape[0] == 1
                res = res[0]

            result_arrays.append(res)
            result_indices.append(i)

        if len(result_arrays) == 0:
            nrows = 0
        else:
            nrows = result_arrays[0].shape[0]
        index = pd.Index(range(nrows))

        columns = self.items

        # error: Argument 1 to "ArrayManager" has incompatible type "List[ndarray]";
        # expected "List[Union[ndarray, ExtensionArray]]"
        return type(self)(result_arrays, [index, columns])  # type: ignore[arg-type]

    def reduce(self, func: Callable[..., Any]) -> EZArrayManager:
        """
        Apply reduction function column-wise, returning a single-row ArrayManager.

        Parameters
        ----------
        func : reduction function

        Returns
        -------
        ArrayManager
        """
        result_arrays: list[npt.NDArray[Any]] = []
        for i, arr in enumerate(self.arrays):
            res = func(arr, axis=0)

            # TODO NaT doesn't preserve dtype, so we need to ensure to create
            # a timedelta result array if original was timedelta
            # what if datetime results in timedelta? (eg std)
            dtype = arr.dtype if res is pd.NaT else None
            result_arrays.append(sanitize_array([res], None, dtype=dtype))

        index = pd.Index._simple_new(np.array([None], dtype=object))  # type: ignore[attr-defined]
        columns = self.items

        # error: Argument 1 to "ArrayManager" has incompatible type "List[ndarray]";
        # expected "List[Union[ndarray, ExtensionArray]]"
        new_mgr = type(self)(result_arrays, [index, columns])  # type: ignore[arg-type]
        return new_mgr

    def operate_blockwise(
        self, other: ArrayManager, array_op: Callable[[zarr.Array, zarr.Array], zarr.Array]
    ) -> ArrayManager:
        """
        Apply array_op blockwise with another (aligned) BlockManager.
        """
        # TODO what if `other` is BlockManager ?
        left_arrays = self.arrays

        if isinstance(other, BlockManager):
            right_arrays = [arr[0] for arr in other.arrays]

        else:
            right_arrays = other.arrays

        assert len(left_arrays) == len(right_arrays)

        result_arrays = [array_op(left, right) for left, right in zip(left_arrays, right_arrays)]
        return ArrayManager(result_arrays, self._axes)

    def quantile(
        self,
        *,
        qs: pd.Index,  # with dtype float64
        transposed: bool = False,
        interpolation: QuantileInterpolation = "linear",
    ) -> ArrayManager:
        arrs = [ensure_block_shape(x, 2) for x in self.arrays]
        new_arrs = [quantile_compat(x, np.asarray(qs._values), interpolation) for x in arrs]  # type: ignore[attr-defined]
        for i, arr in enumerate(new_arrs):
            if arr.ndim == 2:
                assert arr.shape[0] == 1, arr.shape
                new_arrs[i] = arr[0]

        axes = [qs, self._axes[1]]
        return type(self)(new_arrs, axes)

    # ----------------------------------------------------------------
    def unstack(self, unstacker: Any, fill_value: Any) -> EZArrayManager:
        """
        Return a BlockManager with all blocks unstacked.

        Parameters
        ----------
        unstacker : reshape._Unstacker
        fill_value : Any
            fill_value for newly introduced missing values.

        Returns
        -------
        unstacked : BlockManager
        """
        indexer, _ = unstacker._indexer_and_to_sort
        if unstacker.mask.all():
            new_indexer = indexer
            allow_fill = False
            new_mask2D = None
            needs_masking = None
        else:
            new_indexer = np.full(unstacker.mask.shape, -1)
            new_indexer[unstacker.mask] = indexer
            allow_fill = True
            # calculating the full mask once and passing it to take_1d is faster
            # than letting take_1d calculate it in each repeated call
            new_mask2D = (~unstacker.mask).reshape(*unstacker.full_shape)
            needs_masking = new_mask2D.any(axis=0)
        new_indexer2D = new_indexer.reshape(*unstacker.full_shape)
        new_indexer2D = ensure_platform_int(new_indexer2D)

        new_arrays = []
        for arr in self.arrays:
            for i in range(unstacker.full_shape[1]):
                if allow_fill:
                    # error: Value of type "Optional[Any]" is not indexable  [index]
                    new_arr = take_1d(
                        arr,
                        new_indexer2D[:, i],
                        allow_fill=needs_masking[i],  # type: ignore[index]
                        fill_value=fill_value,
                        mask=new_mask2D[:, i],  # type: ignore[index]
                    )
                else:
                    new_arr = take_1d(arr, new_indexer2D[:, i], allow_fill=False)
                new_arrays.append(new_arr)

        new_index = unstacker.new_index
        new_columns = unstacker.get_new_columns(self._axes[1])
        new_axes = [new_index, new_columns]

        return type(self)(new_arrays, new_axes, verify_integrity=False)

    def as_array(self, dtype: Any = None, copy: bool = False, na_value: object = lib.no_default) -> npt.NDArray[Any]:
        """
        Convert the blockmanager data into an numpy array.

        Parameters
        ----------
        dtype : object, default None
            Data type of the return array.
        copy : bool, default False
            If True then guarantee that a copy is returned. A value of
            False does not guarantee that the underlying data is not
            copied.
        na_value : object, default lib.no_default
            Value to be used as the missing value sentinel.

        Returns
        -------
        arr : ndarray
        """
        if len(self.arrays) == 0:
            empty_arr = np.empty(self.shape, dtype=float)
            return empty_arr.transpose()

        # We want to copy when na_value is provided to avoid
        # mutating the original object
        copy = copy or na_value is not lib.no_default

        if not dtype:
            dtype = interleaved_dtype([arr.dtype for arr in self.arrays])

        dtype = ensure_np_dtype(dtype)

        result = np.empty(self.shape_proper, dtype=dtype)

        for i, arr in enumerate(self.arrays):
            arr = arr[:].astype(arr.dtype, copy=copy)
            result[:, i] = arr

        if na_value is not lib.no_default:
            result[isna(result)] = na_value

        return result

    @classmethod
    def concat_horizontal(cls, mgrs: list[EZArrayManager], axes: list[pd.Index]) -> EZArrayManager:
        """
        Concatenate uniformly-indexed ArrayManagers horizontally.
        """
        # concatting along the columns -> combine reindexed arrays in a single manager
        arrays = list(itertools.chain.from_iterable([mgr.arrays for mgr in mgrs]))
        new_mgr = cls(arrays, [axes[1], axes[0]], verify_integrity=False)
        return new_mgr

    @classmethod
    def concat_vertical(cls, mgrs: list[EZArrayManager], axes: list[pd.Index]) -> EZArrayManager:
        """
        Concatenate uniformly-indexed ArrayManagers vertically.
        """
        # concatting along the rows -> concat the reindexed arrays
        # TODO(ArrayManager) doesn't yet preserve the correct dtype
        arrays = [concat_arrays([mgrs[i].arrays[j] for i in range(len(mgrs))]) for j in range(len(mgrs[0].arrays))]
        new_mgr = cls(arrays, [axes[1], axes[0]], verify_integrity=False)
        return new_mgr

    def copy(self, deep: bool | Literal["all"] | None = True) -> ArrayManager:
        if deep is None:
            # ArrayManager does not yet support CoW, so deep=None always means
            # deep=True for now
            deep = True

        if not deep:
            return self

        return ArrayManager([cast(npt.NDArray[Any], arr[:]) for arr in self.arrays], self._axes, verify_integrity=False)  # pyright: ignore[reportInvalidCast]


class EZSingleArrayManager(SingleArrayManager):
    __slots__: list[str] = [
        "_axes",  # private attribute, because 'axes' has different order, see below
        "arrays",
    ]

    arrays: list[zarr.Array]
    _axes: list[pd.Index]

    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self, arrays: list[zarr.Array], axes: list[pd.Index], verify_integrity: bool = True
    ) -> None:
        if verify_integrity:
            assert len(axes) == 1
            assert len(arrays) == 1

            self._axes = [ensure_index(axes[0])]
            self.arrays = [extract_pandas_array(arrays[0])]
            self._verify_integrity()

        else:
            self._axes = axes
            self.arrays = arrays

    def _verify_integrity(self) -> None:
        assert len(self.arrays) == 1

        (n_rows,) = cast(tuple[int], self.shape)
        arr = self.arrays[0]
        assert arr.shape[0] == n_rows

        if not arr.ndim == 1:
            raise ValueError(f"Passed array should be 1-dimensional, got array with {arr.ndim} dimensions instead.")

    # @property
    # @override
    # def ndim(self) -> Literal[1]:
    #     return 1
    #
    # @staticmethod
    # @override
    # def _normalize_axis(axis: Any) -> Any:
    #     return axis
    #
    # @override
    # def make_empty(self, axes: list[pd.Index] | None = None) -> Self:
    #     """Return an empty ArrayManager with index/array of length 0"""
    #     if axes is None:
    #         axes = [pd.Index([], dtype=object)]
    #
    #     array: npt.NDArray[Any] = np.array([], dtype=self.dtype)
    #     raise NotImplementedError
    #     return type(self)([array], axes)

    @classmethod
    def from_array(cls, array: npt.NDArray[Any] | zarr.Array, index: pd.Index) -> SingleArrayManager:
        if not isinstance(array, zarr.Array):
            return SingleArrayManager([array], [index])  # pyright: ignore[reportUnknownVariableType]

        return cls([array], [index])

    @property
    def axes(self) -> list[pd.Index]:
        return self._axes

    @property
    def index(self) -> pd.Index:
        return self._axes[0]

    @property
    def dtype(self) -> np.dtype[Any]:
        return cast(np.dtype[Any], self.array.dtype)

    def external_values(self) -> zarr.Array:
        """The array that Series.values returns"""
        return external_values(self.array)

    def internal_values(self) -> zarr.Array:
        """The array that Series._values returns"""
        return self.array

    # def array_values(self):
    #     """The array that Series.array returns"""
    #     arr = self.array
    #     if isinstance(arr, np.ndarray):
    #         arr = NumpyExtensionArray(arr)
    #     return arr

    @property
    def _can_hold_na(self) -> bool:
        if isinstance(self.array, np.ndarray):
            return self.array.dtype.kind not in "iub"
        else:
            # ExtensionArray
            raise NotImplementedError
            # return self.array._can_hold_na

    @property
    def is_single_block(self) -> bool:
        return True

    def equals(self, other: DataManager) -> bool:
        if not isinstance(other, DataManager):
            return False

        self_axes, other_axes = self.axes, other.axes

        if len(self_axes) != len(other_axes):
            return False

        if not all(ax1.equals(ax2) for ax1, ax2 in zip(self_axes, other_axes)):
            return False

        if not array_equals(self.array, other.array):
            return False

        return True

    def fast_xs(self, loc: int) -> EZSingleArrayManager:
        raise NotImplementedError("Use series._values[loc] instead")

    def get_slice(self, slobj: slice, axis: int = 0) -> EZSingleArrayManager:
        if axis >= self.ndim:
            raise IndexError("Requested axis not found in manager")

        new_array = self.array[slobj]
        new_index = self.index._getitem_slice(slobj)  # type: ignore[attr-defined]
        return type(self)([new_array], [new_index], verify_integrity=False)

    def get_rows_with_mask(self, indexer: npt.NDArray[np.bool_]) -> EZSingleArrayManager:
        new_array = self.array[indexer]
        new_index = self.index[indexer]
        return type(self)([new_array], [new_index])

    def apply(self, func: Callable[[zarr.Array], zarr.Array], **kwargs: Any) -> EZSingleArrayManager:
        if callable(func):
            new_array = func(self.array, **kwargs)
        else:
            new_array = getattr(self.array, func)(**kwargs)
        return type(self)([new_array], self._axes)

    def setitem(self, indexer: Any, value: Any, warn: bool = True) -> EZSingleArrayManager:
        """
        Set values with indexer.

        For SingleArrayManager, this backs s[indexer] = value

        See `setitem_inplace` for a version that works inplace and doesn't
        return a new Manager.
        """
        if isinstance(indexer, np.ndarray) and indexer.ndim > self.ndim:
            raise ValueError(f"Cannot set values with ndim > {self.ndim}")

        return cast(EZSingleArrayManager, self.apply_with_block("setitem", indexer=indexer, value=value))

    def setitem_inplace(self, indexer: Any, value: Any, warn: bool = True) -> None:
        """
        Set values with indexer.

        For Single[Block/Array]Manager, this backs s[indexer] = value

        This is an inplace version of `setitem()`, mutating the manager/values
        in place, not returning a new Manager (and Block), and thus never changing
        the dtype.
        """
        arr = self.array

        # EAs will do this validation in their own __setitem__ methods.
        if isinstance(arr, np.ndarray):
            # Note: checking for ndarray instead of np.dtype means we exclude
            #  dt64/td64, which do their own validation.
            value = np_can_hold_element(arr.dtype, value)

        if isinstance(value, np.ndarray) and value.ndim == 1 and len(value) == 1:
            # NumPy 1.25 deprecation: https://github.com/numpy/numpy/pull/10615
            value = value[0, ...]

        arr[indexer] = value

    def idelete(self, indexer) -> EZSingleArrayManager:
        """
        Delete selected locations in-place (new array, same ArrayManager)
        """
        to_keep = np.ones(self.shape[0], dtype=np.bool_)
        to_keep[indexer] = False

        raise NotImplementedError
        self.arrays = [self.arrays[0][to_keep]]
        self._axes = [self._axes[0][to_keep]]
        return self

    def _reindex_indexer(
        self,
        new_axis: pd.Index,
        indexer: npt.NDArray[np.intp] | None,
        axis: Literal[0],
        fill_value: None = None,
        allow_dups: bool = False,
        copy: bool | None = True,
        use_na_proxy: bool = False,
    ) -> SingleArrayManager:
        """
        Parameters
        ----------
        new_axis : Index
        indexer : ndarray[intp] or None
        axis : int
        fill_value : object, default None
        allow_dups : bool, default False
        copy : bool, default True


        pandas-indexer with -1's only.
        """
        if copy is None:
            # ArrayManager does not yet support CoW, so deep=None always means
            # deep=True for now
            copy = True

        if indexer is None:
            if new_axis is self._axes[axis] and not copy:
                return self

            return type(self)(self.arrays, [new_axis], verify_integrity=False)

        # some axes don't allow reindexing with dups
        if not allow_dups:
            self._axes[axis]._validate_can_reindex(indexer)  # type: ignore[attr-defined]

        if axis >= self.ndim:
            raise IndexError("Requested axis not found in manager")

        validate_indices(indexer, len(self._axes[0]))
        indexer = ensure_platform_int(indexer)
        mask = cast(npt.NDArray[np.bool_], indexer == -1)
        needs_masking = mask.any()

        if copy:
            new_array = take_1d(self.array[:], indexer, allow_fill=needs_masking, fill_value=fill_value, mask=mask)
            return SingleArrayManager([new_array], [new_axis], verify_integrity=False)

        new_array = take_1d(self.array[:], indexer, allow_fill=needs_masking, fill_value=fill_value, mask=mask)
        group = zarr.open_group(self.array.store, path=self.array.name.rstrip(self.array.basename))
        self.arrays = [group.create_array(self.array.basename, data=new_array, overwrite=True)]

        return type(self)(self.arrays, [new_axis], verify_integrity=False)

    def _get_data_subset(self, predicate: Callable[[zarr.Array], bool]) -> EZSingleArrayManager:
        # used in get_numeric_data / get_bool_data
        if predicate(self.array):
            return type(self)(self.arrays, self._axes, verify_integrity=False)

        else:
            return self.make_empty()

    def set_values(self, values: zarr.Array) -> None:
        """
        Set (replace) the values of the SingleArrayManager in place.

        Use at your own risk! This does not check if the passed values are
        valid for the current SingleArrayManager (length, dtype, etc).
        """
        self.arrays[0] = values

    def to_2d_mgr(self, columns: pd.Index) -> EZArrayManager:
        """
        Manager analogue of Series.to_frame
        """
        _array = self.arrays[0]
        arrays = [_array]
        axes = [self.axes[0], columns]

        return EZArrayManager(zarr.open_group(_array), arrays, axes, verify_integrity=False)



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator, MutableSequence
import math
from pathlib import Path
from typing import Any, Self, cast, overload, override
import warnings

import numpy as np
import numpy.typing as npt
import zarr
from numpy._typing import _SupportsArray as SupportsArray  # pyright: ignore[reportPrivateUsage]
from zarr.core.common import AccessModeLiteral
from zarr.errors import BoundsCheckError, UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

from ezarr.common import write_array  # pyright: ignore[reportUnknownVariableType]
from ezarr.names import UNKNOWN, Attribute, EZType
from ezarr.object import EZObject
from ezarr.types import DeferredCreationFunc


def deferred_EZVector(values: Collection[Any]) -> DeferredCreationFunc:
    def inner(grp: zarr.Group, name: str, overwrite: bool) -> None:
        arr = write_array(
            np.array(values),
            grp.store,
            name=name,
            attributes={Attribute.EZType: EZType.Vector, Attribute.EZVectorSize: len(values)},
            overwrite=overwrite,
        )
        arr[:] = values  # pyright: ignore[reportArgumentType]

    return inner


class EZVector[T: np.generic](EZObject[list[T]], MutableSequence[T]):
    """
    List-like object built on top of zarr.Array for storing a single kind of elements in a resizeable container.

    Args:
        array: a zarr.Array, the underlying buffer
        size: (optional) the number of elements stored in the buffer (default: the length of the buffer)

    Example:
        >>> EZVector(zarr.array([1, 2, 3]))
        EZVector[1 2 3]
    """

    def __init__(self, array: zarr.Array, size: int | None = None) -> None:
        assert array.ndim == 1
        self._array: zarr.Array = array

        self._size: int = array.shape[0]
        self._iter_index: int = 0

    @classmethod
    def from_list(
        cls,
        lst: Collection[T],
        *,
        store: StoreLike | None = None,
        name: str,
        path: str | None = None,
        dtype: npt.DTypeLike = None,
        overwrite: bool = False,
    ) -> Self:
        """
        Create an EZVector from a list-like object.

        Args:
            lst: list-like collection of values to store in an EZVector.
            store:
            name:
            path:
            dtype:
            overwrite:
        """
        path = f"{path.rstrip('/')}/{name}" if path else name
        if store is None:
            store = {}

        data = np.asarray(lst)
        if dtype is not None:
            data = data.astype(dtype)

        return cls(write_array(data, store, name=name, overwrite=overwrite))

    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> npt.NDArray[Any]:
        if not copy:
            raise ValueError("converting to a numpy array always creates a copy.")

        data = cast(npt.NDArray[Any], self._array[: self._size])  # pyright: ignore[reportInvalidCast]

        if dtype is not None:
            return data.astype(dtype)

        return data

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}{np.array2string(self._array[: self._size])}"  # pyright: ignore[reportArgumentType]

    @override
    def __len__(self) -> int:
        return self._size

    @overload
    def __getitem__(self, key: int) -> T: ...
    @overload
    def __getitem__(self, key: slice[int] | slice[int, int] | slice[int, int, int]) -> npt.NDArray[T]: ...
    @override
    def __getitem__(self, key: int | slice[int, int, int]) -> T | npt.NDArray[T]:  # pyright: ignore[reportIncompatibleMethodOverride]
        try:
            return cast(npt.NDArray[T], self._array[: self._size])[key]  # pyright: ignore[reportInvalidCast]

        except IndexError:
            raise BoundsCheckError(f"index {key} is out of range for EZVector with size {self._size}")

    @overload
    def __setitem__(self, key: int, value: T) -> None: ...
    @overload
    def __setitem__(self, key: slice[int] | slice[int, int] | slice[int, int, int], value: Iterable[T]) -> None: ...
    @override
    def __setitem__(self, key: int | slice[int, int, int], value: T | Iterable[T]) -> None:
        if isinstance(key, slice):
            self._array[key] = np.asarray(value)
            return

        if key >= self._size:
            raise BoundsCheckError(f"index {key} is out of range for EZVector with size {self._size}")

        self._array[key] = np.array(value)

    @override
    def __delitem__(self, key: int | slice[int, int, int]) -> None:
        """
        Delete a value from this EZVector from an index.

        Args:
            key: value index.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> del ez_vec[1]
            >>> ez_vec
            EZVector[1 3]
        """
        if not isinstance(key, slice):
            key = slice(key, key + 1, None)

        start = 0 if key.start is None else key.start  # pyright: ignore[reportUnnecessaryComparison]
        stop = self._size if key.stop is None else key.stop  # pyright: ignore[reportUnnecessaryComparison]
        step = key.step or 1

        if start < 0:
            start = self._size + start

        if stop < 0:
            stop = self._size + stop

        if step <= 0:
            raise IndexError("Only steps >= 1 are supported")

        if stop > self._size:
            raise IndexError(f"Index {stop} if out of bounds")

        offset = 0
        for idx in range(start, stop, step):
            self._array[idx - offset : -1] = self._array[idx - offset + 1 :]
            offset += 1

        self._size -= offset

    @override
    def __iter__(self) -> Iterator[T]:
        self._iter_index = 0
        return self

    def __next__(self) -> T:
        if self._iter_index >= self._size:
            raise StopIteration

        obj = cast(T, self._array[self._iter_index])
        self._iter_index += 1

        return obj

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, list | SupportsArray):
            return False

        return np.array_equal(self._array[: self._size], other)  # pyright: ignore[reportUnknownArgumentType]

    @property
    def capacity(self) -> int:
        return self._array.shape[0]

    @classmethod
    @override
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open an EZVector from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZVector, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.save("/tmp/vec", overwrite=True)
            >>> EZVector.open("/tmp/vec", name="data")
            EZVector[1 2 3]
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if isinstance(store, str | Path):
            store = Path(store).expanduser()

        arr = zarr.open_array(store, mode=mode, path=path)

        arr_type = cast(str, arr.attrs.get(Attribute.EZType, UNKNOWN))
        if arr_type != EZType.Vector:
            raise ValueError(
                f"Cannot read EZVector from Array marked with type '{arr_type}', should be '{EZType.Vector}'."
            )

        return cls(arr, int(arr.attrs[Attribute.EZVectorSize]))  # pyright: ignore[reportArgumentType]

    @override
    def save(
        self,
        store: StoreLike,
        *,
        name: str | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this EZVector to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZVector, to use inside the store. (default: same as name used during EZVector's creation)
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZVector will be saved.
            overwrite: overwrite EZVector if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.save("/tmp/vec", overwrite=True)
            >>> Path("/tmp/vec").exists()
            True
        """
        if name is None:
            name = self._array.name

        path = f"{path.rstrip('/')}/{name}" if path else name

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            write_array(
                self._array,
                store,
                name=path,
                attributes={Attribute.EZType: EZType.Vector, Attribute.EZVectorSize: self._size},
                overwrite=overwrite,
            )

    @override
    def copy(self) -> list[T]:
        """
        Convert this EZVector into a Python list, loading all the data into memory.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.copy()
            [1, 2, 3]
        """
        return self._array[: self._size].tolist()  # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue]

    @classmethod
    def defer(cls, values: Collection[Any]) -> DeferredCreationFunc:
        r"""
        Create an EZVector from a list-like object, but defer its storage.
        This function returns a function that can be called to finalize the EZVector creation and storage,
        with parameters:
            grp: (zarr.Group) a group in which to store the EZVector
            name: (str) name for the EZVector, to use inside the store.
            overwrite: (bool) overwrite object if a group with name `name` already exists ? (default: False)

        This is usually used as a convenience function to create an EZVector in an existing EZDict.

        Args:
            values: list-like collection of values to store in an EZVector.

        Example:
            >>> from ezarr import EZDict
            >>> d = EZDict(zarr.open_group({}))
            >>> d["vec"] = EZVector.defer([1, 2, 3])
            >>> repr(d)
            'EZDict{\n\tvec: EZVector[1 2 3]\n}'
        """
        return deferred_EZVector(values)

    @override
    def append(self, value: T) -> None:
        """
        Append a value to the end of an EZVector.

        Args:
            value: value to be appended.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.append(4)
            >>> ez_vec
            EZVector[1 2 3 4]
        """
        if self._size == self._array.shape[0]:
            self._array.resize((self._array.shape[0] * 2,))

        self._array[self._size] = value
        self._size += 1

    @override
    def extend(self, values: Iterable[T]) -> None:
        """
        Append multiple values to the end of an EZVector.

        Args:
            values: values to be appended.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.extend([4, 5, 6])
            >>> ez_vec
            EZVector[1 2 3 4 5 6]
        """
        values = list(values)
        n_values = len(values)

        if self._array.shape[0] - self._size < n_values:
            new_size = 2 ** math.ceil(math.log2(self._size + n_values))
            self._array.resize((new_size,))

        self._array[self._size : self._size + n_values] = values
        self._size += n_values

    @override
    def insert(self, index: int, value: T) -> None:
        """
        Insert a value in an EZList before `index`.

        Args:
            value: value to be inserted.

        Example:
            >>> ez_vec = EZVector.from_list([1, 2, 3], name="data")
            >>> ez_vec.insert(1, 4)
            >>> ez_vec
            EZVector[1 4 2 3]
        """
        if index < 0:
            index = self._size + index

        if self._size == self._array.shape[0]:
            self._array.resize((self._array.shape[0] * 2,))

        self._array[index + 1 :] = self._array[index:-1]
        self._array[index] = value
        self._size += 1



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator
from typing import Any, Self, SupportsIndex, cast, overload, override

import numpy as np
import numpy.typing as npt
import zarr
from numpy._typing import _ArrayLikeObject_co  # pyright: ignore[reportPrivateUsage]

from ezarr import indexing
from ezarr.indexing._typing import SELECTOR

_RESERVED_ATTRIBUTES = frozenset(("_array", "_index", "_exposed_attr", "size", "shape", "dtype", "array_type"))


class ArrayView[T](Collection[T]):
    """View on a zarr.Array that can be subsetted infinitely without copying."""

    __slots__: tuple[str, ...] = "_array", "_index", "_exposed_attr"

    def __init__(
        self,
        array: zarr.Array,
        index: SELECTOR | tuple[SELECTOR, ...] | indexing.Selection,
        exposed_attributes: Iterable[str] = (),
    ) -> None:
        self._array: zarr.Array = array
        self._index: indexing.Selection = (
            index if isinstance(index, indexing.Selection) else indexing.Selection.from_selector(index, array.shape)
        )
        self._exposed_attr: frozenset[str] = frozenset(exposed_attributes)

        if not self._exposed_attr.isdisjoint(_RESERVED_ATTRIBUTES):
            raise ValueError(
                f"Cannot expose reserved attributes : {_RESERVED_ATTRIBUTES.intersection(self._exposed_attr)}."
            )

    @override
    def __repr__(self) -> str:
        return repr(self._view()) + "*"

    def __array__(self, dtype: npt.DTypeLike = None, copy: bool | None = None) -> npt.NDArray[Any]:
        if dtype is None:
            return self._view(copy=copy)

        return self._view(copy=copy).astype(dtype)

    @override
    def __getattribute__(self, key: str) -> Any:
        if key in object.__getattribute__(self, "_exposed_attr"):
            return getattr(self._array, key)

        return super().__getattribute__(key)

    @overload
    def __getitem__(self, index: SupportsIndex | tuple[SupportsIndex, ...]) -> T: ...
    @overload
    def __getitem__(
        self,
        index: slice
        | npt.NDArray[np.integer]
        | npt.NDArray[np.bool_]
        | None
        | tuple[slice | npt.NDArray[np.integer] | npt.NDArray[np.bool_] | None, ...],
    ) -> ArrayView[T]: ...
    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> ArrayView[T] | T:
        sel = indexing.Selection.from_selector(index, self.shape).cast_on(self._index)

        if np.prod(sel.out_shape) == 1:
            return cast(T, self._array[sel.get_indexers()])

        return ArrayView(self._array, sel, exposed_attributes=self._exposed_attr)

    def __setitem__(self, index: SELECTOR, values: Any) -> None:
        sel = indexing.Selection.from_selector(index, self.shape).cast_on(self._index)

        self._array[sel.get_indexers()] = values

    @override
    def __len__(self) -> int:
        return len(self._view())

    @override
    def __contains__(self, key: Any) -> bool:
        return key in self._view()

    @override
    def __iter__(self) -> Iterator[T]:
        return iter(self._view())

    @override
    def __eq__(self, __value: object) -> npt.NDArray[np.bool_]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self._view().__eq__(__value)

    def __lt__(self, __value: _ArrayLikeObject_co) -> npt.NDArray[np.bool_]:
        return self._view().__lt__(__value)

    def __le__(self, __value: _ArrayLikeObject_co) -> npt.NDArray[np.bool_]:
        return self._view().__le__(__value)

    def __gt__(self, __value: _ArrayLikeObject_co) -> npt.NDArray[np.bool_]:
        return self._view().__gt__(__value)

    def __ge__(self, __value: _ArrayLikeObject_co) -> npt.NDArray[np.bool_]:
        return self._view().__ge__(__value)

    def __add__(self, other: Any) -> npt.NDArray[Any]:
        return cast(npt.NDArray[Any], self._view() + other)

    def __iadd__(self, other: Any) -> Self:
        index = self._index.get_indexers()
        self._array[index] += other
        return self

    def __sub__(self, other: Any) -> npt.NDArray[Any]:
        return cast(npt.NDArray[Any], self._view() - other)

    def __isub__(self, other: Any) -> Self:
        index = self._index.get_indexers()
        self._array[index] -= other
        return self

    def __mul__(self, other: Any) -> npt.NDArray[Any]:
        return cast(npt.NDArray[Any], self._view() * other)

    def __imul__(self, other: Any) -> Self:
        index = self._index.get_indexers()
        self._array[index] *= other
        return self

    def __truediv__(self, other: Any) -> npt.NDArray[Any]:
        return cast(npt.NDArray[Any], self._view() / other)

    def __itruediv__(self, other: Any) -> Self:
        index = self._index.get_indexers()
        self._array[index] /= other
        return self

    def __pow__(self, other: Any) -> npt.NDArray[Any]:
        return cast(npt.NDArray[Any], self._view() ** other)

    @property
    def size(self) -> int:
        return self._view().size

    @property
    def shape(self) -> tuple[int, ...]:
        return self._view().shape

    @property
    def dtype(self) -> np.dtype[Any]:
        return self._view().dtype

    @property
    def array_type(self) -> type[Any]:
        return type(self._array)

    def _view(self, copy: bool | None = None) -> npt.NDArray[Any]:
        array = cast(npt.NDArray[Any], self._array[self._index.get_indexers()])

        if copy is False:
            raise ValueError("Getting a view of a zarr.Array always creates a copy")

        return array

    def copy(self) -> npt.NDArray[Any]:
        return self._view().copy()

    def astype(self, dtype: npt.DTypeLike) -> npt.NDArray[Any]:
        return self._view().astype(dtype)

    def min(
        self, axis: int | tuple[int, ...] | None = None, out: npt.NDArray[Any] | None = None
    ) -> T | npt.NDArray[Any]:
        return self._view().min(axis=axis, out=out)

    def max(
        self, axis: int | tuple[int, ...] | None = None, out: npt.NDArray[Any] | None = None
    ) -> T | npt.NDArray[Any]:
        return self._view().max(axis=axis, out=out)

    def mean(
        self,
        axis: int | tuple[int, ...] | None = None,
        dtype: npt.DTypeLike | None = None,
        out: npt.NDArray[Any] | None = None,
    ) -> T | npt.NDArray[Any]:
        return self._view().mean(axis=axis, dtype=dtype, out=out)  # type: ignore[no-any-return, misc, arg-type]

    def flatten(self) -> npt.NDArray[Any]:
        return self._view().flatten()

    def tolist(self) -> list[T]:
        return self._view().tolist()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Self

from zarr.core.common import AccessModeLiteral
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]


class EZObject[PythonType](ABC):
    @classmethod
    @abstractmethod
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open this object from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.
        """

    @abstractmethod
    def save(
        self,
        store: StoreLike,
        *,
        name: str,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this object to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the object will be saved.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)
        """

    @abstractmethod
    def copy(self) -> PythonType:
        """
        Convert this object into its Python equivalent, loading all the data into memory.
        """
        pass
